#!/usr/bin/env python3

"""
## File-related tasks

This module provides tasks for manipulating the filesystem.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def download(
    src,
    dest,
    user=None,
    group=None,
    mode=None,
    cache_time=None,
    force=False,
    sha256sum=None,
    sha1sum=None,
    md5sum=None,
    headers=None,
    insecure=False,
):
    """
    Download files from remote locations using ``curl`` or ``wget``.

    + src: source URL of the file
    + dest: where to save the file
    + user: user to own the files
    + group: group to own the files
    + mode: permissions of the files
    + cache_time: if the file exists already, re-download after this time (in seconds)
    + force: always download the file, even if it already exists
    + sha256sum: sha256 hash to checksum the downloaded file against
    + sha1sum: sha1 hash to checksum the downloaded file against
    + md5sum: md5 hash to checksum the downloaded file against
    + headers: optional dictionary of headers to set for the HTTP request
    + insecure: disable SSL verification for the HTTP request

    **Example:**

    ```python
    files.download(
        name="Download the Docker repo file",
        src="https://download.docker.com/linux/centos/docker-ce.repo",
        dest="/etc/yum.repos.d/docker-ce.repo",
    )
    ```
    """
    operargs = {
        "src": repr(src),
        "dest": repr(dest),
        "user": repr(user),
        "group": repr(group),
        "mode": repr(mode),
        "cache_time": repr(cache_time),
        "force": repr(force),
        "sha256sum": repr(sha256sum),
        "sha1sum": repr(sha1sum),
        "md5sum": repr(md5sum),
        "headers": repr(headers),
        "insecure": repr(insecure),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import files", "files.download", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def line(
    path,
    line,
    present=True,
    replace=None,
    flags=None,
    backup=False,
    interpolate_variables=False,
    escape_regex_characters=False,
    assume_present=False,
    ensure_newline=False,
):
    """
    Ensure lines in files using grep to locate and sed to replace.

    + path: target remote file to edit
    + line: string or regex matching the target line
    + present: whether the line should be in the file
    + replace: text to replace entire matching lines when ``present=True``
    + flags: list of flags to pass to sed when replacing/deleting
    + backup: whether to backup the file (see below)
    + interpolate_variables: whether to interpolate variables in ``replace``
    + assume_present: whether to assume a matching line already exists in the file
    + escape_regex_characters: whether to escape regex characters from the matching line
    + ensure_newline: ensures that the appended line is on a new line

    Regex line matching:
        Unless line matches a line (starts with ^, ends $), pyinfra will wrap it such that
        it does, like: ``^.*LINE.*$``. This means we don't swap parts of lines out. To
        change bits of lines, see ``files.replace``.

    Regex line escaping:
        If matching special characters (eg a crontab line containing *), remember to escape
        it first using Python's ``re.escape``.

    Backup:
        If set to ``True``, any editing of the file will place an old copy with the ISO
        date (taken from the machine running ``pyinfra``) appended as the extension. If
        you pass a string value this will be used as the extension of the backed up file.

    Append:
        If ``line`` is not in the file but we want it (``present`` set to ``True``), then
        it will be append to the end of the file.

    Ensure new line:
        This will ensure that the ``line`` being appended is always on a seperate new
        line in case the file doesn't end with a newline character.


    **Examples:**

    ```python
    # prepare to do some maintenance
    maintenance_line = "SYSTEM IS DOWN FOR MAINTENANCE"
    files.line(
        name="Add the down-for-maintenance line in /etc/motd",
        path="/etc/motd",
        line=maintenance_line,
    )

    # Then, after the maintenance is done, remove the maintenance line
    files.line(
        name="Remove the down-for-maintenance line in /etc/motd",
        path="/etc/motd",
        line=maintenance_line,
        replace="",
        present=False,
    )

    # example where there is '*' in the line
    files.line(
        name="Ensure /netboot/nfs is in /etc/exports",
        path="/etc/exports",
        line=r"/netboot/nfs .*",
        replace="/netboot/nfs *(ro,sync,no_wdelay,insecure_locks,no_root_squash,"
        "insecure,no_subtree_check)",
    )

    files.line(
        name="Ensure myweb can run /usr/bin/python3 without password",
        path="/etc/sudoers",
        line=r"myweb .*",
        replace="myweb ALL=(ALL) NOPASSWD: /usr/bin/python3",
    )

    # example when there are double quotes (")
    line = 'QUOTAUSER=""'
    files.line(
        name="Example with double quotes (")",
        path="/etc/adduser.conf",
        line="^{}$".format(line),
        replace=line,
    )
    ```
    """
    operargs = {
        "path": repr(path),
        "line": repr(line),
        "present": repr(present),
        "replace": repr(replace),
        "flags": repr(flags),
        "backup": repr(backup),
        "interpolate_variables": repr(interpolate_variables),
        "escape_regex_characters": repr(escape_regex_characters),
        "assume_present": repr(assume_present),
        "ensure_newline": repr(ensure_newline),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import files", "files.line", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def replace(
    path,
    text=None,
    replace=None,
    flags=None,
    backup=False,
    interpolate_variables=False,
    match=None,
):
    """
    Replace contents of a file using ``sed``.

    + path: target remote file to edit
    + text: text/regex to match against
    + replace: text to replace with
    + flags: list of flags to pass to sed
    + backup: whether to backup the file (see below)
    + interpolate_variables: whether to interpolate variables in ``replace``

    Backup:
        If set to ``True``, any editing of the file will place an old copy with the ISO
        date (taken from the machine running ``pyinfra``) appended as the extension. If
        you pass a string value this will be used as the extension of the backed up file.

    **Example:**

    ```python
    files.replace(
        name="Change part of a line in a file",
        path="/etc/motd",
        text="verboten",
        replace="forbidden",
    )
    ```
    """
    operargs = {
        "path": repr(path),
        "text": repr(text),
        "replace": repr(replace),
        "flags": repr(flags),
        "backup": repr(backup),
        "interpolate_variables": repr(interpolate_variables),
        "match": repr(match),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import files", "files.replace", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def sync(
    src,
    dest,
    user=None,
    group=None,
    mode=None,
    dir_mode=None,
    delete=False,
    exclude=None,
    exclude_dir=None,
    add_deploy_dir=True,
):
    r"""
    Syncs a local directory with a remote one, with delete support. Note that delete will
    remove extra files on the remote side, but not extra directories.

    + src: local directory to sync
    + dest: remote directory to sync to
    + user: user to own the files and directories
    + group: group to own the files and directories
    + mode: permissions of the files
    + dir_mode: permissions of the directories
    + delete: delete remote files not present locally
    + exclude: string or list/tuple of strings to match & exclude files (eg *.pyc)
    + exclude_dir: string or list/tuple of strings to match & exclude directories (eg node_modules)
    + add_deploy_dir: interpret src as relative to deploy directory instead of current directory

    **Example:**

    ```python
    # Sync local files/tempdir to remote /tmp/tempdir
    files.sync(
        name="Sync a local directory with remote",
        src="files/tempdir",
        dest="/tmp/tempdir",
    )
    ```

    Note: ``exclude`` and ``exclude_dir`` use ``fnmatch`` behind the scenes to do the filtering.

    + ``exclude`` matches against the filename.
    + ``exclude_dir`` matches against the path of the directory, relative to ``src``.
      Since fnmatch does not treat path separators (``/`` or ``\``) as special characters,
      excluding all directories matching a given name, however deep under ``src`` they are,
      can be done for example with ``exclude_dir=["__pycache__", "*/__pycache__"]``
    """
    operargs = {
        "src": repr(src),
        "dest": repr(dest),
        "user": repr(user),
        "group": repr(group),
        "mode": repr(mode),
        "dir_mode": repr(dir_mode),
        "delete": repr(delete),
        "exclude": repr(exclude),
        "exclude_dir": repr(exclude_dir),
        "add_deploy_dir": repr(add_deploy_dir),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import files", "files.sync", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def show_rsync_warning():
    operargs = {}

    result = _run_pyinfra(
        "from pyinfra.operations import files", "files.show_rsync_warning", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def rsync(src, dest, flags=["-ax", "--delete"]):
    """
    Use ``rsync`` to sync a local directory to the remote system. This operation will actually call
    the ``rsync`` binary on your system.

    .. important::
        The ``files.rsync`` operation is in alpha, and only supported using SSH
        or ``@local`` connectors. When using the SSH connector, rsync will automatically use the
        StrictHostKeyChecking setting, config and known_hosts file (when specified).

    .. caution::
        When using SSH, the ``files.rsync`` operation only supports the ``sudo`` and ``sudo_user``
        global arguments.
    """
    operargs = {
        "src": repr(src),
        "dest": repr(dest),
        "flags": repr(flags),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import files", "files.rsync", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def _create_remote_dir(state, host, remote_filename, user, group):
    operargs = {
        "state": repr(state),
        "host": repr(host),
        "remote_filename": repr(remote_filename),
        "user": repr(user),
        "group": repr(group),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import files", "files._create_remote_dir", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def get(src, dest, add_deploy_dir=True, create_local_dir=False, force=False):
    """
    Download a file from the remote system.

    + src: the remote filename to download
    + dest: the local filename to download the file to
    + add_deploy_dir: dest is relative to the deploy directory
    + create_local_dir: create the local directory if it doesn't exist
    + force: always download the file, even if the local copy matches

    Note:
        This operation is not suitable for large files as it may involve copying
        the remote file before downloading it.

    **Example:**

    ```python
    files.get(
        name="Download a file from a remote",
        src="/etc/centos-release",
        dest="/tmp/whocares",
    )
    ```
    """
    operargs = {
        "src": repr(src),
        "dest": repr(dest),
        "add_deploy_dir": repr(add_deploy_dir),
        "create_local_dir": repr(create_local_dir),
        "force": repr(force),
    }

    result = _run_pyinfra("from pyinfra.operations import files", "files.get", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def put(
    src,
    dest,
    user=None,
    group=None,
    mode=None,
    add_deploy_dir=True,
    create_remote_dir=True,
    force=False,
    assume_exists=False,
):
    """
    Upload a local file, or file-like object, to the remote system.

    + src: filename or IO-like object to upload
    + dest: remote filename to upload to
    + user: user to own the files
    + group: group to own the files
    + mode: permissions of the files, use ``True`` to copy the local file
    + add_deploy_dir: src is relative to the deploy directory
    + create_remote_dir: create the remote directory if it doesn't exist
    + force: always upload the file, even if the remote copy matches
    + assume_exists: whether to assume the local file exists

    ``dest``:
        If this is a directory that already exists on the remote side, the local
        file will be uploaded to that directory with the same filename.

    ``mode``:
        When set to ``True`` the permissions of the local file are applied to the
        remote file after the upload is complete.

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    Note:
        This operation is not suitable for large files as it may involve copying
        the file before uploading it.

    **Examples:**

    ```python
    files.put(
        name="Update the message of the day file",
        src="files/motd",
        dest="/etc/motd",
        mode="644",
    )

    files.put(
        name="Upload a StringIO object",
        src=StringIO("file contents"),
        dest="/etc/motd",
    )
    ```
    """
    operargs = {
        "src": repr(src),
        "dest": repr(dest),
        "user": repr(user),
        "group": repr(group),
        "mode": repr(mode),
        "add_deploy_dir": repr(add_deploy_dir),
        "create_remote_dir": repr(create_remote_dir),
        "force": repr(force),
        "assume_exists": repr(assume_exists),
    }

    result = _run_pyinfra("from pyinfra.operations import files", "files.put", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def template(
    src, dest, user=None, group=None, mode=None, create_remote_dir=True, **data
):
    """
    Generate a template using jinja2 and write it to the remote system.

    + src: template filename or IO-like object
    + dest: remote filename
    + user: user to own the files
    + group: group to own the files
    + mode: permissions of the files
    + create_remote_dir: create the remote directory if it doesn't exist

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    Notes:
       Common convention is to store templates in a "templates" directory and
       have a filename suffix with '.j2' (for jinja2).

       For information on the template syntax, see
       `the jinja2 docs <https://jinja.palletsprojects.com>`_.

    **Examples:**

    ```python
    files.template(
        name="Create a templated file",
        src="templates/somefile.conf.j2",
        dest="/etc/somefile.conf",
    )

    files.template(
        name="Create service file",
        src="templates/myweb.service.j2",
        dest="/etc/systemd/system/myweb.service",
        mode="755",
        user="root",
        group="root",
    )

    # Example showing how to pass python variable to template file. You can also
    # use dicts and lists. The .j2 file can use `{{ foo_variable }}` to be interpolated.
    foo_variable = 'This is some foo variable contents'
    foo_dict = {
        "str1": "This is string 1",
        "str2": "This is string 2"
    }
    foo_list = [
        "entry 1",
        "entry 2"
    ]

    template = StringIO('''
    name: "{{ foo_variable }}"
    dict_contents:
        str1: "{{ foo_dict.str1 }}"
        str2: "{{ foo_dict.str2 }}"
    list_contents:
    {% for entry in foo_list %}
        - "{{ entry }}"
    {% endfor %}
    ''')

    files.template(
        name="Create a templated file",
        src=template,
        dest="/tmp/foo.yml",
        foo_variable=foo_variable,
        foo_dict=foo_dict,
        foo_list=foo_list
    )
    ```
    """
    operargs = {
        "src": repr(src),
        "dest": repr(dest),
        "user": repr(user),
        "group": repr(group),
        "mode": repr(mode),
        "create_remote_dir": repr(create_remote_dir),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import files", "files.template", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def _validate_path(path):
    operargs = {
        "path": repr(path),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import files", "files._validate_path", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def _raise_or_remove_invalid_path(fs_type, path, force, force_backup, force_backup_dir):
    operargs = {
        "fs_type": repr(fs_type),
        "path": repr(path),
        "force": repr(force),
        "force_backup": repr(force_backup),
        "force_backup_dir": repr(force_backup_dir),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import files",
        "files._raise_or_remove_invalid_path",
        operargs,
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def link(
    path,
    target=None,
    present=True,
    assume_present=False,
    user=None,
    group=None,
    symbolic=True,
    create_remote_dir=True,
    force=False,
    force_backup=True,
    force_backup_dir=None,
):
    """
    Add/remove/update links.

    + path: the name of the link
    + target: the file/directory the link points to
    + present: whether the link should exist
    + assume_present: whether to assume the link exists
    + user: user to own the link
    + group: group to own the link
    + symbolic: whether to make a symbolic link (vs hard link)
    + create_remote_dir: create the remote directory if it doesn't exist
    + force: if the target exists and is not a file, move or remove it and continue
    + force_backup: set to ``False`` to remove any existing non-file when ``force=True``
    + force_backup_dir: directory to move any backup to when ``force=True``

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    Source changes:
        If the link exists and points to a different target, pyinfra will remove it and
        recreate a new one pointing to then new target.

    **Examples:**

    ```python
    files.link(
        name="Create link /etc/issue2 that points to /etc/issue",
        path="/etc/issue2",
        target="/etc/issue",
    )


    # Complex example demonstrating the assume_present option
    from pyinfra.operations import apt, files

    install_nginx = apt.packages(
        name="Install nginx",
        packages=["nginx"],
    )

    files.link(
        name="Remove default nginx site",
        path="/etc/nginx/sites-enabled/default",
        present=False,
        assume_present=install_nginx.changed,
    )
    ```
    """
    operargs = {
        "path": repr(path),
        "target": repr(target),
        "present": repr(present),
        "assume_present": repr(assume_present),
        "user": repr(user),
        "group": repr(group),
        "symbolic": repr(symbolic),
        "create_remote_dir": repr(create_remote_dir),
        "force": repr(force),
        "force_backup": repr(force_backup),
        "force_backup_dir": repr(force_backup_dir),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import files", "files.link", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def file(
    path,
    present=True,
    assume_present=False,
    user=None,
    group=None,
    mode=None,
    touch=False,
    create_remote_dir=True,
    force=False,
    force_backup=True,
    force_backup_dir=None,
):
    """
    Add/remove/update files.

    + path: name/path of the remote file
    + present: whether the file should exist
    + assume_present: whether to assume the file exists
    + user: user to own the files
    + group: group to own the files
    + mode: permissions of the files as an integer, eg: 755
    + touch: whether to touch the file
    + create_remote_dir: create the remote directory if it doesn't exist
    + force: if the target exists and is not a file, move or remove it and continue
    + force_backup: set to ``False`` to remove any existing non-file when ``force=True``
    + force_backup_dir: directory to move any backup to when ``force=True``

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    **Example:**

    ```python
    # Note: The directory /tmp/secret will get created with the default umask.
    files.file(
        name="Create /tmp/secret/file",
        path="/tmp/secret/file",
        mode="600",
        user="root",
        group="root",
        touch=True,
        create_remote_dir=True,
    )
    ```
    """
    operargs = {
        "path": repr(path),
        "present": repr(present),
        "assume_present": repr(assume_present),
        "user": repr(user),
        "group": repr(group),
        "mode": repr(mode),
        "touch": repr(touch),
        "create_remote_dir": repr(create_remote_dir),
        "force": repr(force),
        "force_backup": repr(force_backup),
        "force_backup_dir": repr(force_backup_dir),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import files", "files.file", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def directory(
    path,
    present=True,
    assume_present=False,
    user=None,
    group=None,
    mode=None,
    recursive=False,
    force=False,
    force_backup=True,
    force_backup_dir=None,
    _no_check_owner_mode=False,
    _no_fail_on_link=False,
):
    """
    Add/remove/update directories.

    + path: path of the remote folder
    + present: whether the folder should exist
    + assume_present: whether to assume the directory exists
    + user: user to own the folder
    + group: group to own the folder
    + mode: permissions of the folder
    + recursive: recursively apply user/group/mode
    + force: if the target exists and is not a file, move or remove it and continue
    + force_backup: set to ``False`` to remove any existing non-file when ``force=True``
    + force_backup_dir: directory to move any backup to when ``force=True``

    **Examples:**

    ```python
    files.directory(
        name="Ensure the /tmp/dir_that_we_want_removed is removed",
        path="/tmp/dir_that_we_want_removed",
        present=False,
    )

    files.directory(
        name="Ensure /web exists",
        path="/web",
        user="myweb",
        group="myweb",
    )

    # Multiple directories
    for dir in ["/netboot/tftp", "/netboot/nfs"]:
        files.directory(
            name="Ensure the directory `{}` exists".format(dir),
            path=dir,
        )
    ```
    """
    operargs = {
        "path": repr(path),
        "present": repr(present),
        "assume_present": repr(assume_present),
        "user": repr(user),
        "group": repr(group),
        "mode": repr(mode),
        "recursive": repr(recursive),
        "force": repr(force),
        "force_backup": repr(force_backup),
        "force_backup_dir": repr(force_backup_dir),
        "_no_check_owner_mode": repr(_no_check_owner_mode),
        "_no_fail_on_link": repr(_no_fail_on_link),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import files", "files.directory", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def flags(path, flags=None, present=True):
    """
    Set/clear file flags.

    + path: path of the remote folder
    + flags: a list of the file flags to be set or cleared
    + present: whether the flags should be set or cleared

    **Examples:**

    ```python
    files.flags(
        name="Ensure ~/Library is visible in the GUI",
        path="~/Library",
        flags="hidden",
        present=False
    )

    files.directory(
        name="Ensure no one can change these files",
        path="/something/very/important",
        flags=["uchg", "schg"],
        present=True,
        _sudo=True
    )
    ```
    """
    operargs = {
        "path": repr(path),
        "flags": repr(flags),
        "present": repr(present),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import files", "files.flags", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def block(
    path,
    content=None,
    present=True,
    line=None,
    backup=False,
    escape_regex_characters=False,
    before=False,
    after=False,
    marker=None,
    begin=None,
    end=None,
):
    r"""
    Ensure content, surrounded by the appropriate markers, is present (or not) in the file.

    + path: target remote file
    + content: what should be present in the file (between markers).
    + present: whether the content should be present in the file
    + before: should the content be added before ``line`` if it doesn't exist
    + after: should the content be added after ``line`` if it doesn't exist
    + line: regex before or after which the content should be added if it doesn't exist.
    + backup: whether to backup the file (see ``files.line``). Default False.
    + escape_regex_characters: whether to escape regex characters from the matching line
    + marker: the base string used to mark the text.  Default is ``# {mark} PYINFRA BLOCK``
    + begin: the value for ``{mark}`` in the marker before the content. Default is ``BEGIN``
    + end: the value for ``{mark}`` in the marker after the content. Default is ``END``

    Content appended if ``line`` not found in the file
        If ``content`` is not in the file but is required (``present=True``) and ``line`` is not
        found in the file, ``content`` (surrounded by markers) will be appended to the file.  The
        file is created if necessary.

    Content prepended or appended if ``line`` not specified
        If ``content`` is not in the file but is required and ``line`` was not provided the content
        will either be  prepended to the file (if both ``before`` and ``after``
        are ``True``) or appended to the file (if both are ``False``).

    Removal ignores ``content`` and ``line``

    **Examples:**

    ```python
    # add entry to /etc/host
    files.marked_block(
        name="add IP address for red server",
        path="/etc/hosts",
        content="10.0.0.1 mars-one",
        before=True,
        regex=".*localhost",
    )

    # have two entries in /etc/host
    files.marked_block(
        name="add IP address for red server",
        path="/etc/hosts",
        content="10.0.0.1 mars-one\n10.0.0.2 mars-two",
        before=True,
        regex=".*localhost",
    )

    # remove marked entry from /etc/hosts
    files.marked_block(
        name="remove all 10.* addresses from /etc/hosts",
        path="/etc/hosts",
        present=False
    )

    # add out of date warning to web page
    files.marked_block(
        name="add out of date warning to web page",
        path="/var/www/html/something.html",
        content= "<p>Warning: this page is out of date.</p>",
        regex=".*<body>.*",
        after=True
        marker="<!-- {mark} PYINFRA BLOCK -->",
    )
    ```
    """
    operargs = {
        "path": repr(path),
        "content": repr(content),
        "present": repr(present),
        "line": repr(line),
        "backup": repr(backup),
        "escape_regex_characters": repr(escape_regex_characters),
        "before": repr(before),
        "after": repr(after),
        "marker": repr(marker),
        "begin": repr(begin),
        "end": repr(end),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import files", "files.block", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## Yum package tasks

Manage yum packages and repositories. Note that yum package names are case-sensitive.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def key(src):
    """
    Add yum gpg keys with ``rpm``.

    + src: filename or URL

    Note:
        always returns one command, not state checking

    **Example:**

    .. code:: python

        linux_id = host.get_fact(LinuxDistribution)["release_meta"].get("ID")
        yum.key(
            name="Add the Docker CentOS gpg key",
            src=f"https://download.docker.com/linux/{linux_id}/gpg",
        )
    """
    operargs = {
        "src": repr(src),
    }

    result = _run_pyinfra("from pyinfra.operations import yum", "yum.key", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def repo(
    src,
    present=True,
    baseurl=None,
    description=None,
    enabled=True,
    gpgcheck=True,
    gpgkey=None,
):
    """
    Add/remove/update yum repositories.

    + src: URL or name for the ``.repo``   file
    + present: whether the ``.repo`` file should be present
    + baseurl: the baseurl of the repo (if ``name`` is not a URL)
    + description: optional verbose description
    + enabled: whether this repo is enabled
    + gpgcheck: whether set ``gpgcheck=1``
    + gpgkey: the URL to the gpg key for this repo

    ``Baseurl``/``description``/``gpgcheck``/``gpgkey``:
        These are only valid when ``src`` is a filename (ie not a URL). This is
        for manual construction of repository files. Use a URL to download and
        install remote repository files.

    **Examples:**

    .. code:: python

        # Download a repository file
        yum.repo(
            name="Install Docker-CE repo via URL",
            src="https://download.docker.com/linux/centos/docker-ce.repo",
        )

        # Create the repository file from baseurl/etc
        yum.repo(
            name="Add the Docker CentOS repo",
            src="DockerCE",
            baseurl="https://download.docker.com/linux/centos/7/$basearch/stable",
        )
    """
    operargs = {
        "src": repr(src),
        "present": repr(present),
        "baseurl": repr(baseurl),
        "description": repr(description),
        "enabled": repr(enabled),
        "gpgcheck": repr(gpgcheck),
        "gpgkey": repr(gpgkey),
    }

    result = _run_pyinfra("from pyinfra.operations import yum", "yum.repo", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def rpm(src, present=True):
    """
    Add/remove ``.rpm`` file packages.

    + src: filename or URL of the ``.rpm`` package
    + present: whether ore not the package should exist on the system

    URL sources with ``present=False``:
        If the ``.rpm`` file isn't downloaded, pyinfra can't remove any existing
        package as the file won't exist until mid-deploy.

    **Example:**

    .. code:: python

        major_version = host.get_fact(LinuxDistribution)["major"]
        dnf.rpm(
           name="Install EPEL rpm to enable EPEL repo",
           src=f"https://dl.fedoraproject.org/pub/epel/epel-release-latest-{major_version}.noarch.rpm",
        )
    """
    operargs = {
        "src": repr(src),
        "present": repr(present),
    }

    result = _run_pyinfra("from pyinfra.operations import yum", "yum.rpm", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def update():
    """
    Updates all yum packages.
    """
    operargs = {}

    result = _run_pyinfra("from pyinfra.operations import yum", "yum.update", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def packages(
    packages=None,
    present=True,
    latest=False,
    update=False,
    clean=False,
    nobest=False,
    extra_install_args=None,
    extra_uninstall_args=None,
):
    """
    Install/remove/update yum packages & updates.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version
    + update: run ``yum update`` before installing packages
    + clean: run ``yum clean all`` before installing packages
    + nobest: add the no best option to install
    + extra_install_args: additional arguments to the yum install command
    + extra_uninstall_args: additional arguments to the yum uninstall command

    Versions:
        Package versions can be pinned as follows: ``<pkg>=<version>``

    **Examples:**

    .. code:: python

        # Update package list and install packages
        yum.packages(
            name="Install Vim and Vim enhanced",
            packages=["vim-enhanced", "vim"],
            update=True,
        )

        # Install the latest versions of packages (always check)
        yum.packages(
            name="Install latest Vim",
            packages=["vim"],
            latest=True,
        )
    """
    operargs = {
        "packages": repr(packages),
        "present": repr(present),
        "latest": repr(latest),
        "update": repr(update),
        "clean": repr(clean),
        "nobest": repr(nobest),
        "extra_install_args": repr(extra_install_args),
        "extra_uninstall_args": repr(extra_uninstall_args),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import yum", "yum.packages", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
##  pyinfra Introduction

uPlaybook wrappers of pyinfra https://docs.pyinfra.com/en/2.x/ operators.

This set of tasks provides a rich set of system management routines.
This wraps the pyinfra application.
"""

#  Full docs are in `docs/tasks/pyinfra/intro.md`

from collections import namedtuple
from typing import Dict
import tempfile
import subprocess
import os
import re

PyInfraResults = namedtuple("PyInfraResults", ["changed", "no_change", "errors"])


class PyInfraGlobalArgContext(dict):
    def __init__(self):
        super().__init__(self)


pyinfra_global_args = PyInfraGlobalArgContext()


class PyInfraFailed(Exception):
    def __init__(self, message, stdout, stderr):
        message = f"pyinfra execution failed: {message}\n   stdout: {stdout}\n   stderr: {stderr}"
        super().__init__(message, stdout, stderr)


def _run_pyinfra(
    imports: str, operator: str, operargs: Dict[str, object]
) -> PyInfraResults:
    """
    Run a pyinfra operation.

    Args:
        imports: The imports that the operator will need.
        operator: The name of the operator to run.
        operargs: kwargs-style arguments to the operator, the value needs to be a
                valid python value of the type appropraite for the argument.
    """
    operargs = operargs.copy()
    operargs.update(pyinfra_global_args)

    with tempfile.NamedTemporaryFile(mode="w", suffix=".py", delete=False) as tmp_file:
        tmp_file.write(imports)
        tmp_file.write("\n")
        tmp_file.write(operator)
        tmp_file.write("(")
        for k, v in operargs.items():
            tmp_file.write(f"{k}={v}, ")
        tmp_file.write(")")

        tmp_file.close()

        s = subprocess.run(
            ["pyinfra", "@local", tmp_file.name], text=True, capture_output=True
        )

        os.remove(tmp_file.name)

        if s.returncode != 0:
            raise PyInfraFailed(
                f"Exit code {s.returncode}, expecting 0.", s.stdout, s.stderr
            )

        # [@local]   Changed: 0   No change: 1   Errors: 0
        match = re.search(
            r"\[@local\]\s+Changed:\s*(?P<changed>\d+)\s+No change:\s*(?P<no_change>\d+)\s+Errors:\s*(?P<errors>\d+)",
            s.stderr,
        )
        if not match:
            raise PyInfraFailed(
                f"Unable to parse pyinfra output for 'Changed' message",
                s.stdout,
                s.stderr,
            )

        groups = match.groupdict()
        return PyInfraResults(
            int(groups["changed"]),
            int(groups["no_change"]),
            int(groups["errors"]),
        )


from . import apk
from . import apt
from . import brew
from . import bsdinit
from . import choco
from . import dnf
from . import files
from . import gem
from . import git
from . import iptables
from . import launchd
from . import lxd
from . import mysql
from . import npm
from . import openrc
from . import pacman
from . import pip
from . import pkg
from . import pkgin
from . import postgresql
from . import puppet
from . import server
from . import ssh
from . import systemd
from . import sysvinit
from . import upstart
from . import vzctl
from . import windows
from . import windows_files
from . import xbps
from . import yum
from . import zypper



#!/usr/bin/env python3

"""
## npm


Manage npm (aka node aka Node.js) packages.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def packages(packages=None, present=True, latest=False, directory=None):
    """
    Install/remove/update npm packages.

    + packages: list of packages to ensure
    + present: whether the packages should be present
    + latest: whether to upgrade packages without a specified version
    + directory: directory to manage packages for, defaults to global

    Versions:
        Package versions can be pinned like npm: ``<pkg>@<version>``.
    """
    operargs = {
        "packages": repr(packages),
        "present": repr(present),
        "latest": repr(latest),
        "directory": repr(directory),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import npm", "npm.packages", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## Pip tasks

This module provides tasks for interacting with pip packages.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def virtualenv(
    path, python=None, venv=False, site_packages=False, always_copy=False, present=True
):
    """
    Add/remove Python virtualenvs.

    + python: python interpreter to use
    + venv: use standard venv module instead of virtualenv
    + site_packages: give access to the global site-packages
    + always_copy: always copy files rather than symlinking
    + present: whether the virtualenv should exist

    **Example:**

    .. code:: python

        pip.virtualenv(
            name="Create a virtualenv",
            path="/usr/local/bin/venv",
        )
    """
    operargs = {
        "path": repr(path),
        "python": repr(python),
        "venv": repr(venv),
        "site_packages": repr(site_packages),
        "always_copy": repr(always_copy),
        "present": repr(present),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import pip", "pip.virtualenv", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def venv(path, python=None, site_packages=False, always_copy=False, present=True):
    """
    Add/remove Python virtualenvs.

    + python: python interpreter to use
    + site_packages: give access to the global site-packages
    + always_copy: always copy files rather than symlinking
    + present: whether the virtualenv should exist

    **Example:**

    .. code:: python

        pip.venv(
            name="Create a virtualenv",
            path="/usr/local/bin/venv",
        )
    """
    operargs = {
        "path": repr(path),
        "python": repr(python),
        "site_packages": repr(site_packages),
        "always_copy": repr(always_copy),
        "present": repr(present),
    }

    result = _run_pyinfra("from pyinfra.operations import pip", "pip.venv", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def packages(
    packages=None,
    present=True,
    latest=False,
    requirements=None,
    pip="pip3",
    virtualenv=None,
    virtualenv_kwargs=None,
    extra_install_args=None,
):
    """
    Install/remove/update pip packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version
    + requirements: location of requirements file to install/uninstall
    + pip: name or path of the pip directory to use
    + virtualenv: root directory of virtualenv to work in
    + virtualenv_kwargs: dictionary of arguments to pass to ``pip.virtualenv``
    + extra_install_args: additional arguments to the pip install command

    Virtualenv:
        This will be created if it does not exist already. ``virtualenv_kwargs``
        will be passed to ``pip.virtualenv`` which can be used to control how
        the env is created.

    Versions:
        Package versions can be pinned like pip: ``<pkg>==<version>``.

    **Example:**

    .. code:: python

        pip.packages(
            name="Install pyinfra into a virtualenv",
            packages=["pyinfra"],
            virtualenv="/usr/local/bin/venv",
        )
    """
    operargs = {
        "packages": repr(packages),
        "present": repr(present),
        "latest": repr(latest),
        "requirements": repr(requirements),
        "pip": repr(pip),
        "virtualenv": repr(virtualenv),
        "virtualenv_kwargs": repr(virtualenv_kwargs),
        "extra_install_args": repr(extra_install_args),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import pip", "pip.packages", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## windows_files


The windows_files module handles windows filesystem state, file uploads and template generation.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def download(
    src,
    dest,
    user=None,
    group=None,
    mode=None,
    cache_time=None,
    force=False,
    sha256sum=None,
    sha1sum=None,
    md5sum=None,
):
    r"""
    Download files from remote locations using curl or wget.

    + src: source URL of the file
    + dest: where to save the file
    + user: user to own the files
    + group: group to own the files
    + mode: permissions of the files
    + cache_time: if the file exists already, re-download after this time (in seconds)
    + force: always download the file, even if it already exists
    + sha256sum: sha256 hash to checksum the downloaded file against
    + sha1sum: sha1 hash to checksum the downloaded file against
    + md5sum: md5 hash to checksum the downloaded file against

    **Example:**

    .. code:: python

        windows_files.download(
            name="Download the Docker repo file",
            src="https://download.docker.com/linux/centos/docker-ce.repo",
            dest="C:\docker",
        )
    """
    operargs = {
        "src": repr(src),
        "dest": repr(dest),
        "user": repr(user),
        "group": repr(group),
        "mode": repr(mode),
        "cache_time": repr(cache_time),
        "force": repr(force),
        "sha256sum": repr(sha256sum),
        "sha1sum": repr(sha1sum),
        "md5sum": repr(md5sum),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import windows_files",
        "windows_files.download",
        operargs,
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def put(
    src,
    dest,
    user=None,
    group=None,
    mode=None,
    add_deploy_dir=True,
    create_remote_dir=True,
    force=False,
    assume_exists=False,
):
    r"""
    Upload a local file to the remote system.

    + src: local filename to upload
    + dest: remote filename to upload to
    + user: user to own the files
    + group: group to own the files
    + mode: permissions of the files
    + add_deploy_dir: src is relative to the deploy directory
    + create_remote_dir: create the remote directory if it doesn't exist
    + force: always upload the file, even if the remote copy matches
    + assume_exists: whether to assume the local file exists

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    Note:
        This operation is not suitable for large files as it may involve copying
        the file before uploading it.

    **Examples:**

    .. code:: python

        # Note: This requires a 'files/motd' file on the local filesystem
        files.put(
            name="Update the message of the day file",
            src="data/content.json",
            dest="C:\data\content.json",
        )
    """
    operargs = {
        "src": repr(src),
        "dest": repr(dest),
        "user": repr(user),
        "group": repr(group),
        "mode": repr(mode),
        "add_deploy_dir": repr(add_deploy_dir),
        "create_remote_dir": repr(create_remote_dir),
        "force": repr(force),
        "assume_exists": repr(assume_exists),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import windows_files", "windows_files.put", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def file(
    path,
    present=True,
    assume_present=False,
    user=None,
    group=None,
    mode=None,
    touch=False,
    create_remote_dir=True,
):
    r"""
    Add/remove/update files.

    + path: path of the remote file
    + present: whether the file should exist
    + assume_present: whether to assume the file exists
    + TODO: user: user to own the files
    + TODO: group: group to own the files
    + TODO: mode: permissions of the files as an integer, eg: 755
    + touch: whether to touch the file
    + create_remote_dir: create the remote directory if it doesn't exist

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    **Example:**

    .. code:: python

        files.file(
            name="Create c:\temp\hello.txt",
            path="c:\temp\hello.txt",
            touch=True,
        )
    """
    operargs = {
        "path": repr(path),
        "present": repr(present),
        "assume_present": repr(assume_present),
        "user": repr(user),
        "group": repr(group),
        "mode": repr(mode),
        "touch": repr(touch),
        "create_remote_dir": repr(create_remote_dir),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import windows_files", "windows_files.file", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def _create_remote_dir(state, host, remote_filename, user, group):
    operargs = {
        "state": repr(state),
        "host": repr(host),
        "remote_filename": repr(remote_filename),
        "user": repr(user),
        "group": repr(group),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import windows_files",
        "windows_files._create_remote_dir",
        operargs,
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def directory(
    path,
    present=True,
    assume_present=False,
    user=None,
    group=None,
    mode=None,
    recursive=False,
):
    r"""
    Add/remove/update directories.

    + path: path of the remote folder
    + present: whether the folder should exist
    + assume_present: whether to assume the directory exists
    + TODO: user: user to own the folder
    + TODO: group: group to own the folder
    + TODO: mode: permissions of the folder
    + TODO: recursive: recursively apply user/group/mode

    **Examples:**

    .. code:: python

        files.directory(
            name="Ensure the c:\temp\dir_that_we_want_removed is removed",
            path="c:\temp\dir_that_we_want_removed",
            present=False,
        )

        files.directory(
            name="Ensure c:\temp\foo\foo_dir exists",
            path="c:\temp\foo\foo_dir",
            recursive=True,
        )

        # multiple directories
        dirs = ["c:\temp\foo_dir1", "c:\temp\foo_dir2"]
        for dir in dirs:
            files.directory(
                name="Ensure the directory `{}` exists".format(dir),
                path=dir,
            )
    """
    operargs = {
        "path": repr(path),
        "present": repr(present),
        "assume_present": repr(assume_present),
        "user": repr(user),
        "group": repr(group),
        "mode": repr(mode),
        "recursive": repr(recursive),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import windows_files",
        "windows_files.directory",
        operargs,
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def _validate_path(path):
    operargs = {
        "path": repr(path),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import windows_files",
        "windows_files._validate_path",
        operargs,
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def link(
    path,
    target=None,
    present=True,
    assume_present=False,
    user=None,
    group=None,
    symbolic=True,
    force=True,
    create_remote_dir=True,
):
    r"""
    Add/remove/update links.

    + path: the name of the link
    + target: the file/directory the link points to
    + present: whether the link should exist
    + assume_present: whether to assume the link exists
    + user: user to own the link
    + group: group to own the link
    + symbolic: whether to make a symbolic link (vs hard link)
    + create_remote_dir: create the remote directory if it doesn't exist

    ``create_remote_dir``:
        If the remote directory does not exist it will be created using the same
        user & group as passed to ``files.put``. The mode will *not* be copied over,
        if this is required call ``files.directory`` separately.

    Source changes:
        If the link exists and points to a different target, pyinfra will remove it and
        recreate a new one pointing to then new target.

    **Examples:**

    .. code:: python

        # simple example showing how to link to a file
        files.link(
            name=r"Create link C:\issue2 that points to C:\issue",
            path=r"C:\issue2",
            target=r"C\issue",
        )
    """
    operargs = {
        "path": repr(path),
        "target": repr(target),
        "present": repr(present),
        "assume_present": repr(assume_present),
        "user": repr(user),
        "group": repr(group),
        "symbolic": repr(symbolic),
        "force": repr(force),
        "create_remote_dir": repr(create_remote_dir),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import windows_files", "windows_files.link", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
A module for handling transparently reading encrypted source files or teplates.
"""

#  from fernetcrypt
encrypt_blocksize = 40960
decrypt_blocksize = 54712
salt_file_size = 20
magic = "#UF1#"

from cryptography.fernet import Fernet
from cryptography.hazmat.primitives import hashes
from cryptography.hazmat.primitives.kdf.pbkdf2 import PBKDF2HMAC
import base64
from typing import Union, Callable, Dict

from jinja2 import Template


class FernetReader:
    """
    A file-like class to transparently read FernetCrypt encrypted files, falling back
    to reading plain-text files if the file is not encrypted.

    Args:
        - file_name: File to read.
        - password: Can either be a password, or a function that returns a password when
            called.

    Examples:

        def get_password():
            return "foo"


        f = FernetReader("foo.fernet", "foo")
        print(f.read())
        f = FernetReader("foo.fernet", get_password)
        print(f.read())

        with FernetReader("foo.fernet", get_password) as f:
           print(f.read())
    """

    def __init__(self, file_name: str, password: Union[str, Callable]):
        self.file_name = file_name
        self.password = password

        self._file = open(self.file_name, "rb")
        self.buffer = self._file.read(25)
        if self.buffer[20:25] == b"#UF1#":
            password = self._get_password()
            salt = base64.b85decode(self.buffer[:20])

            kdf = PBKDF2HMAC(
                algorithm=hashes.SHA256(),
                length=32,
                salt=salt,
                iterations=960000,
            )
            key = base64.urlsafe_b64encode(kdf.derive(password.encode("ascii")))
            self._fernet = Fernet(key)
            self._reader = self.fernet_reader

            self.buffer = b""
        else:
            self._reader = self.plaintext_reader

    def _get_password(self):
        """Returns the password for decryption."""

        if callable(self.password):
            return self.password()
        return self.password

    def fernet_reader(self, size: int = -1) -> bytes:
        while True:
            #  already have enough bytes
            if size != -1 and len(self.buffer) > size:
                ret = self.buffer[:size]
                self.buffer = self.buffer[size:]
                return ret

            data = self._file.read(decrypt_blocksize)
            if not data:
                ret = self.buffer
                self.buffer = b""
                return ret

            self.buffer += self._fernet.decrypt(data)

    def plaintext_reader(self, size: int = -1) -> bytes:
        ret = self.buffer + self._file.read(size)
        self.buffer = b""
        return ret

    def read(self, size: int = -1):
        return self._reader(size)

    def __enter__(self):
        """Enter the runtime context related to this object."""
        return self

    def __exit__(self, *_):
        """Exit the runtime context and perform any necessary cleanup."""
        self._file.close()


class JinjaFernetReader:
    """
    A wrapper on top of FernetReader that will render the template contained in the file.

    Args:
        - file_name: File to read.
        - password: Can either be a password, or a function that returns a password when
            called.
        - context: Jinja context to use with rendering
        - encoding: The encoding to convert the source file binary data into str.

    Examples:

        f = JinaFernetReader("foo.fernet", "foo", {"foo": "bar"})
        print(f.read())
    """

    def __init__(
        self,
        file_name: str,
        password: Union[str, Callable],
        context: Dict[str, str],
        encoding: str = "latin-1",
    ):
        with FernetReader(file_name, password) as fp:
            self.generator = Template(fp.read().decode(encoding)).generate(context)
        self.buffer = ""

    def read(self, size: int = -1) -> str:
        while size < 0 or len(self.buffer) < size:
            try:
                self.buffer += next(self.generator)
            except StopIteration:
                break
        if size < 0:
            ret = self.buffer
            self.buffer = ""
        else:
            ret = self.buffer[:size]
            self.buffer = self.buffer[size:]

        return ret

    def __enter__(self):
        """Enter the runtime context related to this object."""
        return self

    def __exit__(self, *_):
        """Exit the runtime context and perform any necessary cleanup."""
        pass



#!/usr/bin/env python3

"""
## Git tasks

This module provides tasks for interfacing with git version control.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def config(key, value, multi_value=False, repo=None):
    """
    Manage git config for a repository or globally.

    + key: the key of the config to ensure
    + value: the value this key should have
    + multi_value: Add the value rather than set it for settings that can have multiple values
    + repo: specify the git repo path to edit local config (defaults to global)

    **Example:**

    .. code:: python

        git.config(
            name="Ensure user name is set for a repo",
            key="user.name",
            value="Anon E. Mouse",
            repo="/usr/local/src/pyinfra",
        )
    """
    operargs = {
        "key": repr(key),
        "value": repr(value),
        "multi_value": repr(multi_value),
        "repo": repr(repo),
    }

    result = _run_pyinfra("from pyinfra.operations import git", "git.config", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def repo(
    src,
    dest,
    branch=None,
    pull=True,
    rebase=False,
    user=None,
    group=None,
    ssh_keyscan=False,
    update_submodules=False,
    recursive_submodules=False,
):
    """
    Clone/pull git repositories.

    + src: the git source URL
    + dest: directory to clone to
    + branch: branch to pull/checkout
    + pull: pull any changes for the branch
    + rebase: when pulling, use ``--rebase``
    + user: chown files to this user after
    + group: chown files to this group after
    + ssh_keyscan: keyscan the remote host if not in known_hosts before clone/pull
    + update_submodules: update any git submodules
    + recursive_submodules: update git submodules recursively

    **Example:**

    .. code:: python

        git.repo(
            name="Clone repo",
            src="https://github.com/Fizzadar/pyinfra.git",
            dest="/usr/local/src/pyinfra",
        )
    """
    operargs = {
        "src": repr(src),
        "dest": repr(dest),
        "branch": repr(branch),
        "pull": repr(pull),
        "rebase": repr(rebase),
        "user": repr(user),
        "group": repr(group),
        "ssh_keyscan": repr(ssh_keyscan),
        "update_submodules": repr(update_submodules),
        "recursive_submodules": repr(recursive_submodules),
    }

    result = _run_pyinfra("from pyinfra.operations import git", "git.repo", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def worktree(
    worktree,
    repo=None,
    detached=False,
    new_branch=None,
    commitish=None,
    pull=True,
    rebase=False,
    from_remote_branch=None,
    present=True,
    assume_repo_exists=False,
    force=False,
    user=None,
    group=None,
):
    """
    Manage git worktrees.

    + worktree: git working tree directory
    + repo: git main repository directory
    + detached: create a working tree with a detached HEAD
    + new_branch: local branch name created at the same time than the worktree
    + commitish: from which git commit, branch, ... the worktree is created
    + pull: pull any changes from a remote branch if set
    + rebase: when pulling, use ``--rebase``
    + from_remote_branch: a 2-tuple ``(remote, branch)`` that identifies a remote branch
    + present: whether the working tree should exist
    + assume_repo_exists: whether to assume the main repo exists
    + force: remove unclean working tree if should not exist
    + user: chown files to this user after
    + group: chown files to this group after

    **Examples:**

    .. code:: python

        git.worktree(
            name="Create a worktree from the current repo `HEAD`",
            repo="/usr/local/src/pyinfra/master",
            worktree="/usr/local/src/pyinfra/hotfix"
        )

        git.worktree(
            name="Create a worktree from the commit `4e091aa0`",
            repo="/usr/local/src/pyinfra/master",
            worktree="/usr/local/src/pyinfra/hotfix",
            commitish="4e091aa0"
        )

        git.worktree(
            name="Create a worktree with a new local branch `v1.0`",
            repo="/usr/local/src/pyinfra/master",
            worktree="/usr/local/src/pyinfra/hotfix",
            new_branch="v1.0",
        )

        git.worktree(
            name="Create a worktree from the commit 4e091aa0 with the new local branch `v1.0`",
            repo="/usr/local/src/pyinfra/master",
            worktree="/usr/local/src/pyinfra/hotfix",
            new_branch="v1.0",
            commitish="4e091aa0"
        )

        git.worktree(
            name="Create a worktree with a detached `HEAD`",
            repo="/usr/local/src/pyinfra/master",
            worktree="/usr/local/src/pyinfra/hotfix",
            detached=True,
        )

        git.worktree(
            name="Create a worktree with a detached `HEAD` from commit `4e091aa0`",
            repo="/usr/local/src/pyinfra/master",
            worktree="/usr/local/src/pyinfra/hotfix",
            commitish="4e091aa0",
            detached=True,
        )

        git.worktree(
            name="Create a worktree from the existing local branch `v1.0`",
            repo="/usr/local/src/pyinfra/master",
            worktree="/usr/local/src/pyinfra/hotfix",
            commitish="v1.0"
        )

        git.worktree(
            name="Create a worktree with a new branch `v1.0` that tracks `origin/v1.0`",
            repo="/usr/local/src/pyinfra/master",
            worktree="/usr/local/src/pyinfra/hotfix",
            new_branch="v1.0",
            commitish="v1.0"
        )

        git.worktree(
            name="Pull an existing worktree already linked to a tracking branch",
            repo="/usr/local/src/pyinfra/master",
            worktree="/usr/local/src/pyinfra/hotfix"
        )

        git.worktree(
            name="Pull an existing worktree from a specific remote branch",
            repo="/usr/local/src/pyinfra/master",
            worktree="/usr/local/src/pyinfra/hotfix",
            from_remote_branch=("origin", "master")
        )

        git.worktree(
            name="Remove a worktree",
            worktree="/usr/local/src/pyinfra/hotfix",
            present=False,
        )

        git.worktree(
            name="Remove an unclean worktree",
            worktree="/usr/local/src/pyinfra/hotfix",
            present=False,
            force=True,
        )
    """
    operargs = {
        "worktree": repr(worktree),
        "repo": repr(repo),
        "detached": repr(detached),
        "new_branch": repr(new_branch),
        "commitish": repr(commitish),
        "pull": repr(pull),
        "rebase": repr(rebase),
        "from_remote_branch": repr(from_remote_branch),
        "present": repr(present),
        "assume_repo_exists": repr(assume_repo_exists),
        "force": repr(force),
        "user": repr(user),
        "group": repr(group),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import git", "git.worktree", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def bare_repo(path, user=None, group=None, present=True):
    """
    Create bare git repositories.

    + path: path to the folder
    + present: whether the bare repository should exist
    + user: chown files to this user after
    + group: chown files to this group after

    **Example:**

    .. code:: python

        git.bare_repo(
            name="Create bare repo",
            path="/home/git/test.git",
        )
    """
    operargs = {
        "path": repr(path),
        "user": repr(user),
        "group": repr(group),
        "present": repr(present),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import git", "git.bare_repo", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## Apk package tasks

This module provides tasks for interacting with the apk package manager.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def upgrade(available=False):
    """
    Upgrades all apk packages.

    + available: force all packages to be upgraded (recommended on whole Alpine version upgrades)
    """
    operargs = {
        "available": repr(available),
    }

    result = _run_pyinfra("from pyinfra.operations import apk", "apk.upgrade", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def update():
    """
    Updates apk repositories.
    """
    operargs = {}

    result = _run_pyinfra("from pyinfra.operations import apk", "apk.update", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def packages(packages=None, present=True, latest=False, update=False, upgrade=False):
    """
    Add/remove/update apk packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version
    + update: run ``apk update`` before installing packages
    + upgrade: run ``apk upgrade`` before installing packages

    Versions:
        Package versions can be pinned like apk: ``<pkg>=<version>``.

    **Examples:**

    .. code:: python

        # Update package list and install packages
        apk.packages(
            name="Install Asterisk and Vim",
            packages=["asterisk", "vim"],
            update=True,
        )

        # Install the latest versions of packages (always check)
        apk.packages(
            name="Install latest Vim",
            packages=["vim"],
            latest=True,
        )
    """
    operargs = {
        "packages": repr(packages),
        "present": repr(present),
        "latest": repr(latest),
        "update": repr(update),
        "upgrade": repr(upgrade),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import apk", "apk.packages", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## upstart


Manage upstart services.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def service(
    service, running=True, restarted=False, reloaded=False, command=None, enabled=None
):
    """
    Manage the state of upstart managed services.

    + service: name of the service to manage
    + running: whether the service should be running
    + restarted: whether the service should be restarted
    + reloaded: whether the service should be reloaded
    + command: custom command to pass like: ``/etc/rc.d/<service> <command>``
    + enabled: whether this service should be enabled/disabled on boot

    Enabling/disabling services:
        Upstart jobs define runlevels in their config files - as such there is no way to
        edit/list these without fiddling with the config. So pyinfra simply manages the
        existence of a ``/etc/init/<service>.override`` file, and sets its content to
        "manual" to disable automatic start of services.
    """
    operargs = {
        "service": repr(service),
        "running": repr(running),
        "restarted": repr(restarted),
        "reloaded": repr(reloaded),
        "command": repr(command),
        "enabled": repr(enabled),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import upstart", "upstart.service", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## iptables tasks

This module provides tasks for manipulating the system firewall.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def chain(chain, present=True, table=filter, policy=None, version=4):
    """
    Add/remove/update iptables chains.

    + chain: the name of the chain
    + present: whether the chain should exist
    + table: the iptables table this chain should belong to
    + policy: the policy this table should have
    + version: whether to target iptables or ip6tables

    Policy:
        These can only be applied to system chains (FORWARD, INPUT, OUTPUT, etc).
    """
    operargs = {
        "chain": repr(chain),
        "present": repr(present),
        "table": repr(table),
        "policy": repr(policy),
        "version": repr(version),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import iptables", "iptables.chain", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def rule(
    chain,
    jump,
    present=True,
    table="filter",
    append=True,
    version=4,
    protocol=None,
    not_protocol=None,
    source=None,
    not_source=None,
    destination=None,
    not_destination=None,
    in_interface=None,
    not_in_interface=None,
    out_interface=None,
    not_out_interface=None,
    to_destination=None,
    to_source=None,
    to_ports=None,
    log_prefix=None,
    destination_port=None,
    source_port=None,
    extras="",
):
    """
    Add/remove iptables rules.

    + chain: the chain this rule should live in
    + jump: the target of the rule
    + table: the iptables table this rule should belong to
    + append: whether to append or insert the rule (if not present)
    + version: whether to target iptables or ip6tables

    Iptables args:

    + protocol/not_protocol: filter by protocol (tcp or udp)
    + source/not_source: filter by source IPs
    + destination/not_destination: filter by destination IPs
    + in_interface/not_in_interface: filter by incoming interface
    + out_interface/not_out_interface: filter by outgoing interface
    + to_destination: where to route to when jump=DNAT
    + to_source: where to route to when jump=SNAT
    + to_ports: where to route to when jump=REDIRECT
    + log_prefix: prefix for the log of this rule when jump=LOG

    Extras:

    + extras: a place to define iptables extension arguments (eg --limit, --physdev)
    + destination_port: destination port (requires protocol)
    + source_port: source port (requires protocol)

    **Examples:**

    .. code:: python

        iptables.rule(
            name="Block SSH traffic",
            chain="INPUT",
            jump="DROP",
            destination_port=22,
        )

        iptables.rule(
            name="NAT traffic on from 8.8.8.8:53 to 8.8.4.4:8080",
            chain="PREROUTING",
            jump="DNAT",
            table="nat",
            source="8.8.8.8",
            destination_port=53,
            to_destination="8.8.4.4:8080",
        )
    """
    operargs = {
        "chain": repr(chain),
        "jump": repr(jump),
        "present": repr(present),
        "table": repr(table),
        "append": repr(append),
        "version": repr(version),
        "protocol": repr(protocol),
        "not_protocol": repr(not_protocol),
        "source": repr(source),
        "not_source": repr(not_source),
        "destination": repr(destination),
        "not_destination": repr(not_destination),
        "in_interface": repr(in_interface),
        "not_in_interface": repr(not_in_interface),
        "out_interface": repr(out_interface),
        "not_out_interface": repr(not_out_interface),
        "to_destination": repr(to_destination),
        "to_source": repr(to_source),
        "to_ports": repr(to_ports),
        "log_prefix": repr(log_prefix),
        "destination_port": repr(destination_port),
        "source_port": repr(source_port),
        "extras": repr(extras),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import iptables", "iptables.rule", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## choco


Manage ``choco`` (Chocolatey) packages (https://chocolatey.org).
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def packages(packages=None, present=True, latest=False):
    """
    Add/remove/update ``choco`` packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version

    Versions:
        Package versions can be pinned like gem: ``<pkg>:<version>``.

    **Example:**

    .. code:: python

        # Note: Assumes that 'choco' is installed and
        #       user has Administrator permission.
        choco.packages(
            name="Install Notepad++",
            packages=["notepadplusplus"],
        )
    """
    operargs = {
        "packages": repr(packages),
        "present": repr(present),
        "latest": repr(latest),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import choco", "choco.packages", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def install():
    """
    Install ``choco`` (Chocolatey).
    """
    operargs = {}

    result = _run_pyinfra(
        "from pyinfra.operations import choco", "choco.install", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## pkgin


Manage pkgin packages.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def upgrade():
    """
    Upgrades all pkgin packages.
    """
    operargs = {}

    result = _run_pyinfra(
        "from pyinfra.operations import pkgin", "pkgin.upgrade", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def update():
    """
    Updates pkgin repositories.
    """
    operargs = {}

    result = _run_pyinfra(
        "from pyinfra.operations import pkgin", "pkgin.update", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def packages(packages=None, present=True, latest=False, update=False, upgrade=False):
    """
    Add/remove/update pkgin packages.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version
    + update: run ``pkgin update`` before installing packages
    + upgrade: run ``pkgin upgrade`` before installing packages

    **Examples:**

    .. code:: python

        # Update package list and install packages
        pkgin.packages(
            name="Install tmux and Vim",
            packages=["tmux", "vim"],
            update=True,
        )

        # Install the latest versions of packages (always check)
        pkgin.packages(
            name="Install latest Vim",
            packages=["vim"],
            latest=True,
        )
    """
    operargs = {
        "packages": repr(packages),
        "present": repr(present),
        "latest": repr(latest),
        "update": repr(update),
        "upgrade": repr(upgrade),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import pkgin", "pkgin.packages", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## Mysql tasks

This module provides tasks for working with mysql databases.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def sql(
    sql,
    database=None,
    mysql_user=None,
    mysql_password=None,
    mysql_host=None,
    mysql_port=None,
):
    """
    Execute arbitrary SQL against MySQL.

    + sql: SQL command(s) to execute
    + database: optional database to open the connection with
    + mysql_*: global module arguments, see above
    """
    operargs = {
        "sql": repr(sql),
        "database": repr(database),
        "mysql_user": repr(mysql_user),
        "mysql_password": repr(mysql_password),
        "mysql_host": repr(mysql_host),
        "mysql_port": repr(mysql_port),
    }

    result = _run_pyinfra("from pyinfra.operations import mysql", "mysql.sql", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def user(
    user,
    present=True,
    user_hostname="localhost",
    password=None,
    privileges=None,
    require=None,
    require_cipher=False,
    require_issuer=False,
    require_subject=False,
    max_connections=None,
    max_queries_per_hour=None,
    max_updates_per_hour=None,
    max_connections_per_hour=None,
    mysql_user=None,
    mysql_password=None,
    mysql_host=None,
    mysql_port=None,
):
    """
    Add/remove/update MySQL users.

    + user: the name of the user
    + present: whether the user should exist or not
    + user_hostname: the hostname of the user
    + password: the password of the user (if created)
    + privileges: the global privileges for this user
    + mysql_*: global module arguments, see above

    Hostname:
        this + ``name`` makes the username - so changing this will create a new
        user, rather than update users with the same ``name``.

    Password:
        will only be applied if the user does not exist - ie pyinfra cannot
        detect if the current password doesn't match the one provided, so won't
        attempt to change it.

    **Example:**

    .. code:: python

        mysql.user(
            name="Create the pyinfra@localhost MySQL user",
            user="pyinfra",
            password="somepassword",
        )

        # Create a user with resource limits
        mysql.user(
            name="Create the pyinfra@localhost MySQL user",
            user="pyinfra",
            max_connections=50,
            max_updates_per_hour=10,
        )

        # Create a user that requires SSL for connections
        mysql.user(
            name="Create the pyinfra@localhost MySQL user",
            user="pyinfra",
            password="somepassword",
            require="SSL",
        )

        # Create a user that requires a specific certificate
        mysql.user(
            name="Create the pyinfra@localhost MySQL user",
            user="pyinfra",
            password="somepassword",
            require="X509",
            require_issuer="/C=SE/ST=Stockholm...",
            require_cipher="EDH-RSA-DES-CBC3-SHA",
        )
    """
    operargs = {
        "user": repr(user),
        "present": repr(present),
        "user_hostname": repr(user_hostname),
        "password": repr(password),
        "privileges": repr(privileges),
        "require": repr(require),
        "require_cipher": repr(require_cipher),
        "require_issuer": repr(require_issuer),
        "require_subject": repr(require_subject),
        "max_connections": repr(max_connections),
        "max_queries_per_hour": repr(max_queries_per_hour),
        "max_updates_per_hour": repr(max_updates_per_hour),
        "max_connections_per_hour": repr(max_connections_per_hour),
        "mysql_user": repr(mysql_user),
        "mysql_password": repr(mysql_password),
        "mysql_host": repr(mysql_host),
        "mysql_port": repr(mysql_port),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import mysql", "mysql.user", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def database(
    database,
    present=True,
    collate=None,
    charset=None,
    user=None,
    user_hostname="localhost",
    user_privileges="ALL",
    mysql_user=None,
    mysql_password=None,
    mysql_host=None,
    mysql_port=None,
):
    """
    Add/remove MySQL databases.

    + name: the name of the database
    + present: whether the database should exist or not
    + collate: the collate to use when creating the database
    + charset: the charset to use when creating the database
    + user: MySQL user to grant privileges on this database to
    + user_hostname: the hostname of the MySQL user to grant
    + user_privileges: privileges to grant to any specified user
    + mysql_*: global module arguments, see above

    Collate/charset:
        these will only be applied if the database does not exist - ie pyinfra
        will not attempt to alter the existing databases collate/character sets.

    **Example:**

    .. code:: python

        mysql.database(
            name="Create the pyinfra_stuff database",
            database="pyinfra_stuff",
            user="pyinfra",
            user_privileges=["SELECT", "INSERT"],
            charset="utf8",
        )
    """
    operargs = {
        "database": repr(database),
        "present": repr(present),
        "collate": repr(collate),
        "charset": repr(charset),
        "user": repr(user),
        "user_hostname": repr(user_hostname),
        "user_privileges": repr(user_privileges),
        "mysql_user": repr(mysql_user),
        "mysql_password": repr(mysql_password),
        "mysql_host": repr(mysql_host),
        "mysql_port": repr(mysql_port),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import mysql", "mysql.database", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def privileges(
    user,
    privileges,
    user_hostname="localhost",
    database="*",
    table="*",
    flush=True,
    with_grant_option=False,
    mysql_user=None,
    mysql_password=None,
    mysql_host=None,
    mysql_port=None,
):
    """
    Add/remove MySQL privileges for a user, either global, database or table specific.

    + user: name of the user to manage privileges for
    + privileges: list of privileges the user should have (see also: ``with_grant_option`` argument)
    + user_hostname: the hostname of the user
    + database: name of the database to grant privileges to (defaults to all)
    + table: name of the table to grant privileges to (defaults to all)
    + flush: whether to flush (and update) the privileges table after any changes
    + with_grant_option: whether the grant option privilege should be set
    + mysql_*: global module arguments, see above
    """
    operargs = {
        "user": repr(user),
        "privileges": repr(privileges),
        "user_hostname": repr(user_hostname),
        "database": repr(database),
        "table": repr(table),
        "flush": repr(flush),
        "with_grant_option": repr(with_grant_option),
        "mysql_user": repr(mysql_user),
        "mysql_password": repr(mysql_password),
        "mysql_host": repr(mysql_host),
        "mysql_port": repr(mysql_port),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import mysql", "mysql.privileges", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def dump(
    dest,
    database=None,
    mysql_user=None,
    mysql_password=None,
    mysql_host=None,
    mysql_port=None,
):
    """
    Dump a MySQL database into a ``.sql`` file. Requires ``mysqldump``.

    + dest: name of the file to dump the SQL to
    + database: name of the database to dump
    + mysql_*: global module arguments, see above

    **Example:**

    .. code:: python

        mysql.dump(
            name="Dump the pyinfra_stuff database",
            dest="/tmp/pyinfra_stuff.dump",
            database="pyinfra_stuff",
        )
    """
    operargs = {
        "dest": repr(dest),
        "database": repr(database),
        "mysql_user": repr(mysql_user),
        "mysql_password": repr(mysql_password),
        "mysql_host": repr(mysql_host),
        "mysql_port": repr(mysql_port),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import mysql", "mysql.dump", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def load(
    src,
    database=None,
    mysql_user=None,
    mysql_password=None,
    mysql_host=None,
    mysql_port=None,
):
    """
    Load ``.sql`` file into a database.

    + src: the filename to read from
    + database: name of the database to import into
    + mysql_*: global module arguments, see above

    **Example:**

    .. code:: python

        mysql.load(
            name="Import the pyinfra_stuff dump into pyinfra_stuff_copy",
            src="/tmp/pyinfra_stuff.dump",
            database="pyinfra_stuff_copy",
        )
    """
    operargs = {
        "src": repr(src),
        "database": repr(database),
        "mysql_user": repr(mysql_user),
        "mysql_password": repr(mysql_password),
        "mysql_host": repr(mysql_host),
        "mysql_port": repr(mysql_port),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import mysql", "mysql.load", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def handle_privileges(action, target, privileges_to_apply, with_statement=None):
    operargs = {
        "action": repr(action),
        "target": repr(target),
        "privileges_to_apply": repr(privileges_to_apply),
        "with_statement": repr(with_statement),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import mysql", "mysql.handle_privileges", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def sql(
    sql,
    database=None,
    mysql_user=None,
    mysql_password=None,
    mysql_host=None,
    mysql_port=None,
):
    """
    Execute arbitrary SQL against MySQL.

    + sql: SQL command(s) to execute
    + database: optional database to open the connection with
    + mysql_*: global module arguments, see above
    """
    operargs = {
        "sql": repr(sql),
        "database": repr(database),
        "mysql_user": repr(mysql_user),
        "mysql_password": repr(mysql_password),
        "mysql_host": repr(mysql_host),
        "mysql_port": repr(mysql_port),
    }

    result = _run_pyinfra("from pyinfra.operations import mysql", "mysql.sql", operargs)

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def user(
    user,
    present=True,
    user_hostname="localhost",
    password=None,
    privileges=None,
    require=None,
    require_cipher=False,
    require_issuer=False,
    require_subject=False,
    max_connections=None,
    max_queries_per_hour=None,
    max_updates_per_hour=None,
    max_connections_per_hour=None,
    mysql_user=None,
    mysql_password=None,
    mysql_host=None,
    mysql_port=None,
):
    """
    Add/remove/update MySQL users.

    + user: the name of the user
    + present: whether the user should exist or not
    + user_hostname: the hostname of the user
    + password: the password of the user (if created)
    + privileges: the global privileges for this user
    + mysql_*: global module arguments, see above

    Hostname:
        this + ``name`` makes the username - so changing this will create a new
        user, rather than update users with the same ``name``.

    Password:
        will only be applied if the user does not exist - ie pyinfra cannot
        detect if the current password doesn't match the one provided, so won't
        attempt to change it.

    **Example:**

    .. code:: python

        mysql.user(
            name="Create the pyinfra@localhost MySQL user",
            user="pyinfra",
            password="somepassword",
        )

        # Create a user with resource limits
        mysql.user(
            name="Create the pyinfra@localhost MySQL user",
            user="pyinfra",
            max_connections=50,
            max_updates_per_hour=10,
        )

        # Create a user that requires SSL for connections
        mysql.user(
            name="Create the pyinfra@localhost MySQL user",
            user="pyinfra",
            password="somepassword",
            require="SSL",
        )

        # Create a user that requires a specific certificate
        mysql.user(
            name="Create the pyinfra@localhost MySQL user",
            user="pyinfra",
            password="somepassword",
            require="X509",
            require_issuer="/C=SE/ST=Stockholm...",
            require_cipher="EDH-RSA-DES-CBC3-SHA",
        )
    """
    operargs = {
        "user": repr(user),
        "present": repr(present),
        "user_hostname": repr(user_hostname),
        "password": repr(password),
        "privileges": repr(privileges),
        "require": repr(require),
        "require_cipher": repr(require_cipher),
        "require_issuer": repr(require_issuer),
        "require_subject": repr(require_subject),
        "max_connections": repr(max_connections),
        "max_queries_per_hour": repr(max_queries_per_hour),
        "max_updates_per_hour": repr(max_updates_per_hour),
        "max_connections_per_hour": repr(max_connections_per_hour),
        "mysql_user": repr(mysql_user),
        "mysql_password": repr(mysql_password),
        "mysql_host": repr(mysql_host),
        "mysql_port": repr(mysql_port),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import mysql", "mysql.user", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def database(
    database,
    present=True,
    collate=None,
    charset=None,
    user=None,
    user_hostname="localhost",
    user_privileges="ALL",
    mysql_user=None,
    mysql_password=None,
    mysql_host=None,
    mysql_port=None,
):
    """
    Add/remove MySQL databases.

    + name: the name of the database
    + present: whether the database should exist or not
    + collate: the collate to use when creating the database
    + charset: the charset to use when creating the database
    + user: MySQL user to grant privileges on this database to
    + user_hostname: the hostname of the MySQL user to grant
    + user_privileges: privileges to grant to any specified user
    + mysql_*: global module arguments, see above

    Collate/charset:
        these will only be applied if the database does not exist - ie pyinfra
        will not attempt to alter the existing databases collate/character sets.

    **Example:**

    .. code:: python

        mysql.database(
            name="Create the pyinfra_stuff database",
            database="pyinfra_stuff",
            user="pyinfra",
            user_privileges=["SELECT", "INSERT"],
            charset="utf8",
        )
    """
    operargs = {
        "database": repr(database),
        "present": repr(present),
        "collate": repr(collate),
        "charset": repr(charset),
        "user": repr(user),
        "user_hostname": repr(user_hostname),
        "user_privileges": repr(user_privileges),
        "mysql_user": repr(mysql_user),
        "mysql_password": repr(mysql_password),
        "mysql_host": repr(mysql_host),
        "mysql_port": repr(mysql_port),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import mysql", "mysql.database", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def privileges(
    user,
    privileges,
    user_hostname="localhost",
    database="*",
    table="*",
    flush=True,
    with_grant_option=False,
    mysql_user=None,
    mysql_password=None,
    mysql_host=None,
    mysql_port=None,
):
    """
    Add/remove MySQL privileges for a user, either global, database or table specific.

    + user: name of the user to manage privileges for
    + privileges: list of privileges the user should have (see also: ``with_grant_option`` argument)
    + user_hostname: the hostname of the user
    + database: name of the database to grant privileges to (defaults to all)
    + table: name of the table to grant privileges to (defaults to all)
    + flush: whether to flush (and update) the privileges table after any changes
    + with_grant_option: whether the grant option privilege should be set
    + mysql_*: global module arguments, see above
    """
    operargs = {
        "user": repr(user),
        "privileges": repr(privileges),
        "user_hostname": repr(user_hostname),
        "database": repr(database),
        "table": repr(table),
        "flush": repr(flush),
        "with_grant_option": repr(with_grant_option),
        "mysql_user": repr(mysql_user),
        "mysql_password": repr(mysql_password),
        "mysql_host": repr(mysql_host),
        "mysql_port": repr(mysql_port),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import mysql", "mysql.privileges", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def dump(
    dest,
    database=None,
    mysql_user=None,
    mysql_password=None,
    mysql_host=None,
    mysql_port=None,
):
    """
    Dump a MySQL database into a ``.sql`` file. Requires ``mysqldump``.

    + dest: name of the file to dump the SQL to
    + database: name of the database to dump
    + mysql_*: global module arguments, see above

    **Example:**

    .. code:: python

        mysql.dump(
            name="Dump the pyinfra_stuff database",
            dest="/tmp/pyinfra_stuff.dump",
            database="pyinfra_stuff",
        )
    """
    operargs = {
        "dest": repr(dest),
        "database": repr(database),
        "mysql_user": repr(mysql_user),
        "mysql_password": repr(mysql_password),
        "mysql_host": repr(mysql_host),
        "mysql_port": repr(mysql_port),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import mysql", "mysql.dump", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def load(
    src,
    database=None,
    mysql_user=None,
    mysql_password=None,
    mysql_host=None,
    mysql_port=None,
):
    """
    Load ``.sql`` file into a database.

    + src: the filename to read from
    + database: name of the database to import into
    + mysql_*: global module arguments, see above

    **Example:**

    .. code:: python

        mysql.load(
            name="Import the pyinfra_stuff dump into pyinfra_stuff_copy",
            src="/tmp/pyinfra_stuff.dump",
            database="pyinfra_stuff_copy",
        )
    """
    operargs = {
        "src": repr(src),
        "database": repr(database),
        "mysql_user": repr(mysql_user),
        "mysql_password": repr(mysql_password),
        "mysql_host": repr(mysql_host),
        "mysql_port": repr(mysql_port),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import mysql", "mysql.load", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def handle_privileges(action, target, privileges_to_apply, with_statement=None):
    operargs = {
        "action": repr(action),
        "target": repr(target),
        "privileges_to_apply": repr(privileges_to_apply),
        "with_statement": repr(with_statement),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import mysql", "mysql.handle_privileges", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## launchd


Manage launchd services.
"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def service(service, running=True, restarted=False, command=None):
    """
    Manage the state of systemd managed services.

    + service: name of the service to manage
    + running: whether the service should be running
    + restarted: whether the service should be restarted
    + command: custom command to pass like: ``launchctl <command> <service>``
    + enabled: whether this service should be enabled/disabled on boot
    """
    operargs = {
        "service": repr(service),
        "running": repr(running),
        "restarted": repr(restarted),
        "command": repr(command),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import launchd", "launchd.service", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)



#!/usr/bin/env python3

"""
## zypper


"""

from . import _run_pyinfra, PyInfraFailed, PyInfraResults
from typing import Optional, List
from ..internals import task, TemplateStr, Return


@task
def repo(
    src,
    baseurl=None,
    present=True,
    description=None,
    enabled=True,
    gpgcheck=True,
    gpgkey=None,
    type="rpm-md",
):
    """
    Add/remove/update zypper repositories.

    + src: URL or name for the ``.repo``   file
    + baseurl: the baseurl of the repo (if ``name`` is not a URL)
    + present: whether the ``.repo`` file should be present
    + description: optional verbose description
    + enabled: whether this repo is enabled
    + gpgcheck: whether set ``gpgcheck=1``
    + gpgkey: the URL to the gpg key for this repo
    + type: the type field this repo (defaults to ``rpm-md``)

    ``Baseurl``/``description``/``gpgcheck``/``gpgkey``:
        These are only valid when ``name`` is a filename (ie not a URL). This is
        for manual construction of repository files. Use a URL to download and
        install remote repository files.

    **Examples:**

    .. code:: python

        # Download a repository file
        zypper.repo(
            name="Install container virtualization repo via URL",
            src="https://download.opensuse.org/repositories/Virtualization:containers/openSUSE_Tumbleweed/Virtualization:containers.repo",
        )

        # Create the repository file from baseurl/etc
        zypper.repo(
            name="Install container virtualization repo",
            src=="Virtualization:containers (openSUSE_Tumbleweed)",
            baseurl="https://download.opensuse.org/repositories/Virtualization:/containers/openSUSE_Tumbleweed/",
        )
    """
    operargs = {
        "src": repr(src),
        "baseurl": repr(baseurl),
        "present": repr(present),
        "description": repr(description),
        "enabled": repr(enabled),
        "gpgcheck": repr(gpgcheck),
        "gpgkey": repr(gpgkey),
        "type": repr(type),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import zypper", "zypper.repo", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def rpm(src, present=True):
    """
    Add/remove ``.rpm`` file packages.

    + src: filename or URL of the ``.rpm`` package
    + present: whether ore not the package should exist on the system

    URL sources with ``present=False``:
        If the ``.rpm`` file isn't downloaded, pyinfra can't remove any existing
        package as the file won't exist until mid-deploy.

    **Example:**

    .. code:: python

        zypper.rpm(
           name="Install task from rpm",
           src="https://github.com/go-task/task/releases/download/v2.8.1/task_linux_amd64.rpm",
        )
    """
    operargs = {
        "src": repr(src),
        "present": repr(present),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import zypper", "zypper.rpm", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def update():
    """
    Updates all zypper packages.
    """
    operargs = {}

    result = _run_pyinfra(
        "from pyinfra.operations import zypper", "zypper.update", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)


@task
def packages(
    packages=None,
    present=True,
    latest=False,
    update=False,
    clean=False,
    extra_global_install_args=None,
    extra_install_args=None,
    extra_global_uninstall_args=None,
    extra_uninstall_args=None,
):
    """
    Install/remove/update zypper packages & updates.

    + packages: list of packages to ensure
    + present: whether the packages should be installed
    + latest: whether to upgrade packages without a specified version
    + update: run ``zypper update`` before installing packages
    + clean: run ``zypper clean --all`` before installing packages
    + extra_global_install_args: additional global arguments to the zypper install command
    + extra_install_args: additional arguments to the zypper install command
    + extra_global_uninstall_args: additional global arguments to the zypper uninstall command
    + extra_uninstall_args: additional arguments to the zypper uninstall command

    Versions:
        Package versions can be pinned like zypper: ``<pkg>=<version>``

    **Examples:**

    .. code:: python

        # Update package list and install packages
        zypper.packages(
            name="Install Vim and Vim enhanced",
            packages=["vim-enhanced", "vim"],
            update=True,
        )

        # Install the latest versions of packages (always check)
        zypper.packages(
            name="Install latest Vim",
            packages=["vim"],
            latest=True,
        )
    """
    operargs = {
        "packages": repr(packages),
        "present": repr(present),
        "latest": repr(latest),
        "update": repr(update),
        "clean": repr(clean),
        "extra_global_install_args": repr(extra_global_install_args),
        "extra_install_args": repr(extra_install_args),
        "extra_global_uninstall_args": repr(extra_global_uninstall_args),
        "extra_uninstall_args": repr(extra_uninstall_args),
    }

    result = _run_pyinfra(
        "from pyinfra.operations import zypper", "zypper.packages", operargs
    )

    if result.errors:
        return Return(failure=True)
    return Return(changed=result.changed != 0)
